from ..core import WesternCalendar


class EuropeanCentralBank(WesternCalendar):
    "European Central Bank"
    # Civil holidays
    include_labour_day = True
    FIXED_HOLIDAYS = WesternCalendar.FIXED_HOLIDAYS + (
        (12, 26, "St. Stephen's Day"),
    )

    # Christian holidays
    include_good_friday = True
    include_easter_monday = True



from ..registry_tools import iso_register
from .core import UnitedStates


@iso_register('US-NM')
class NewMexico(UnitedStates):
    """New Mexico"""
    include_thanksgiving_friday = True
    thanksgiving_friday_label = "Presidents' Day"
    include_federal_presidents_day = False



from .core import UnitedStates
from ..registry_tools import iso_register


@iso_register('US-GU')
class Guam(UnitedStates):
    """Guam"""
    FIXED_HOLIDAYS = UnitedStates.FIXED_HOLIDAYS + (
        (3, 7, 'Guam History and Chamorro Heritage Day'),
        (7, 21, 'Liberation Day'),
    )
    include_all_souls = True
    include_immaculate_conception = True
    immaculate_conception_label = "Lady of Camarin Day"



from datetime import date

from ..core import (
    WesternMixin, IslamicMixin, ChineseNewYearCalendar,
    SAT, SUN
)
from ..registry_tools import iso_register


@iso_register('SG')
class Singapore(WesternMixin, IslamicMixin, ChineseNewYearCalendar):
    "Singapore"
    # Civil holidays
    include_labour_day = True

    # Christian holiday
    include_good_friday = True

    # Islamic holidays
    include_eid_al_fitr = True
    eid_al_fitr_label = "Hari Raya Puasa"
    include_day_of_sacrifice = True
    day_of_sacrifice_label = "Hari Raya Haji"

    FIXED_HOLIDAYS = ChineseNewYearCalendar.FIXED_HOLIDAYS + (
        (8, 9, "National Day"),
    )

    # Explicitly assign these WE days, Singapore calendar is too much of a mix
    WEEKEND_DAYS = (SAT, SUN)

    # Diwali/Deepavali is sometimes celebrated on a different day to India
    # so this can't be put into a HinduMixin
    # Ref:
    # https://holidays.kayaposoft.com/public_holidays.php?year=2022&country=sgp
    DEEPAVALI = {
        2000: date(2000, 10, 26),
        2001: date(2001, 11, 14),
        2002: date(2002, 11, 3),
        2003: date(2003, 10, 23),
        2004: date(2004, 11, 11),
        2005: date(2005, 11, 1),
        2006: date(2006, 10, 21),
        2007: date(2007, 11, 8),
        2008: date(2008, 10, 27),
        2009: date(2009, 10, 17),
        2010: date(2010, 11, 5),
        2011: date(2011, 10, 26),
        2012: date(2012, 11, 13),
        2013: date(2013, 11, 3),
        2014: date(2014, 10, 22),
        2015: date(2015, 11, 10),
        2016: date(2016, 10, 29),
        2017: date(2017, 10, 18),
        2018: date(2018, 11, 6),
        2019: date(2019, 10, 27),
        2020: date(2020, 11, 14),
        2021: date(2021, 11, 4),
        2022: date(2022, 10, 24),
        2023: date(2023, 11, 13),
        2024: date(2024, 10, 31),
        2025: date(2025, 10, 20),
        2026: date(2026, 11, 9),
        2027: date(2027, 10, 28),
        2028: date(2028, 11, 15),
        2029: date(2029, 11, 5),
        2030: date(2030, 10, 25),
    }
    chinese_new_year_label = "Chinese Lunar New Year's Day"
    include_chinese_second_day = True
    chinese_second_day_label = "Second day of Chinese Lunar New Year"
    shift_sunday_holidays = True

    def get_variable_days(self, year):
        """
        Singapore variable days
        """
        if year not in self.DEEPAVALI:
            msg = f'Missing date for Singapore Deepavali for year: {year}'
            raise KeyError(msg)

        days = super().get_variable_days(year)

        # Vesak Day
        days.append(
            (ChineseNewYearCalendar.lunar(year, 4, 15), "Vesak Day"),
        )

        # Add in Deepavali (hardcoded dates, so no need to shift)
        days.append((self.DEEPAVALI.get(year), 'Deepavali'))
        return days



from ..core import FRI
from ..registry_tools import iso_register
from .core import UnitedStates


@iso_register('US-NE')
class Nebraska(UnitedStates):
    """Nebraska"""
    include_thanksgiving_friday = True

    def get_variable_days(self, year):
        days = super().get_variable_days(year)
        days.append(
            (self.get_last_weekday_in_month(year, 4, FRI), "Arbor Day")
        )
        return days



from ..registry_tools import iso_register
from .core import UnitedStates


@iso_register('US-WI')
class Wisconsin(UnitedStates):
    """Wisconsin"""
    include_columbus_day = False
    include_federal_presidents_day = False
    include_christmas_eve = True
    FIXED_HOLIDAYS = UnitedStates.FIXED_HOLIDAYS + (
        (12, 31, "New Years Eve"),
    )



from .core import UnitedStates
from ..registry_tools import iso_register


@iso_register('US-ID')
class Idaho(UnitedStates):
    """Idaho"""
    martin_luther_king_label = (
        "Martin Luther King Jr. / Idaho Human Rights Day"
    )



try:
    import skyfield  # noqa: F401
    import skyfield_data  # noqa: F401
    from .skyfield_astronomy import calculate_equinoxes, solar_term
except ImportError:
    from .precomputed_astronomy import calculate_equinoxes, solar_term

__all__ = [
    'calculate_equinoxes',
    'solar_term',
]



from datetime import date
from ..core import OrthodoxCalendar
from ..registry_tools import iso_register


@iso_register('UA')
class Ukraine(OrthodoxCalendar):
    'Ukraine'

    shift_sunday_holidays = True
    shift_new_years_day = True

    FIXED_HOLIDAYS = OrthodoxCalendar.FIXED_HOLIDAYS + (
        (3, 8, "International Women’s Day"),
        (5, 9, "Victory Day"),
    )
    # Civil holidays
    include_labour_day = True
    labour_day_label = "Workers Solidarity Day"
    # Christian holidays
    include_christmas = False
    include_good_friday = True
    include_easter_sunday = True
    include_easter_monday = True
    include_whit_monday = True

    def get_variable_days(self, year):
        days = super().get_variable_days(year)

        # Orthodox Christmas holiday is moved when it falls over the week
        orthodox_christmas = date(year, 1, 7)
        if orthodox_christmas.weekday() in self.get_weekend_days():
            days.append((
                self.find_following_working_day(orthodox_christmas),
                "Orthodox Christmas (postponed)"))
        else:
            days.append((orthodox_christmas, "Orthodox Christmas"))

        # Constitution Day was celebrated for the first time in 1996
        if year >= 1996:
            constitution_day = date(year, 6, 28)
            if constitution_day.weekday() in self.get_weekend_days():
                days.append((
                    self.find_following_working_day(constitution_day),
                    "Constitution Day (postponed)"))
            else:
                days.append((constitution_day, "Constitution Day"))

        # Independence Day was first celebrated in 1991 on the 16th of June
        # After the official independence has been moved to the 24th of August
        if year == 1991:
            days.append((date(year, 6, 16), "Independence Day"))
        if year >= 1992:
            independence_day = date(year, 8, 24)

            if independence_day.weekday() in self.get_weekend_days():
                days.append((
                    self.find_following_working_day(independence_day),
                    "Independence Day (postponed)"))
            else:
                days.append((independence_day, "Independence Day"))

        # Defender of Ukraine from 2015
        # https://en.wikipedia.org/wiki/Defender_of_Ukraine_Day
        if year >= 2015:
            days.append((date(year, 10, 14), "Day of Defender of Ukraine"))

        # Catholic Christmas has become an holiday only starting from 2017
        if year >= 2017:
            days.append((date(year, 12, 25), "Christmas Day"))

        # Workers Solidarity Day was celebrated also on the 2nd till 2017
        if year <= 2017:
            days.append((date(year, 5, 2), "Workers Solidarity Day"))

        return days



from ..registry_tools import iso_register
from .core import UnitedStates


@iso_register('US-OH')
class Ohio(UnitedStates):
    """Ohio"""

# NOTE: Ohio includes only Federal holidays.
# The wikipedia page say it also includes Election Day, but no official
# document confirms this.



from unittest import TestCase

from ..africa import (
    Algeria,
    Benin,
    IvoryCoast,
    Madagascar,
    SaoTomeAndPrincipe,
    SouthAfrica,
)
from ..registry import registry


class RegistryAfrica(TestCase):
    def test_africa(self):
        classes = set(registry.region_registry.values())
        self.assertIn(Algeria, classes)
        self.assertIn(Benin, classes)
        self.assertIn(IvoryCoast, classes)
        self.assertIn(Madagascar, classes)
        self.assertIn(SaoTomeAndPrincipe, classes)
        self.assertIn(SouthAfrica, classes)



import itertools
import functools
from datetime import date, timedelta

from more_itertools import recipes
from dateutil import relativedelta as rd


class Holiday(date):
    """
    A named holiday with an indicated date, name, and additional keyword
    attributes.

    >>> nyd = Holiday(date(2014, 1, 1), "New year")

    But if New Year's Eve is also a holiday, and it too falls on a weekend,
    many calendars will have that holiday fall back to the previous friday:

    >>> from dateutil import relativedelta as rd
    >>> nye = Holiday(date(2014, 12, 31), "New year's eve",
    ...     observance_shift=dict(weekday=rd.FR(-1)))

    For compatibility, a Holiday may be treated like a tuple of (date, label)

    >>> nyd[0] == date(2014, 1, 1)
    True
    >>> nyd[1]
    'New year'
    >>> d, label = nyd
    """

    def __new__(cls, date, *args, **kwargs):
        return super().__new__(
            cls, date.year, date.month, date.day)

    def __init__(self, date, name='Holiday', **kwargs):
        self.name = name
        vars(self).update(kwargs)

    def __getitem__(self, n):
        """
        for compatibility as a two-tuple
        """
        tp = self, self.name
        return tp[n]

    def __iter__(self):
        """
        for compatibility as a two-tuple
        """
        tp = self, self.name
        return iter(tp)

    def replace(self, *args, **kwargs):
        replaced = super().replace(*args, **kwargs)
        vars(replaced).update(vars(self))
        return replaced

    def __add__(self, other):
        orig = date(self.year, self.month, self.day)
        return Holiday(orig + other, **vars(self))

    def __sub__(self, other):
        orig = date(self.year, self.month, self.day)
        return Holiday(orig - other, **vars(self))

    def nearest_weekday(self, calendar):
        """
        Return the nearest weekday to self.
        """
        weekend_days = calendar.get_weekend_days()
        deltas = (timedelta(n) for n in itertools.count())
        candidates = recipes.flatten(
            (self - delta, self + delta)
            for delta in deltas
        )
        matches = (
            day for day in candidates
            if day.weekday() not in weekend_days
        )
        return next(matches)

    @classmethod
    def _from_fixed_definition(cls, item):
        """For backward compatibility, load Holiday object from an item of
        FIXED_HOLIDAYS class property, which might be just a tuple of
        month, day, label.
        """
        if isinstance(item, tuple):
            month, day, label = item
            any_year = 2000
            item = Holiday(date(any_year, month, day), label)
        return item

    @classmethod
    def _from_resolved_definition(cls, item, **kwargs):
        """For backward compatibility, load Holiday object from a two-tuple
        or existing Holiday instance.
        """
        if isinstance(item, tuple):
            item = Holiday(*item, **kwargs)
        return item

    @functools.lru_cache()
    def get_observed_date(self, calendar):
        """
        The date the holiday is observed for the calendar. If the holiday
        occurs on a weekend, it may be observed on another day as indicated by
        the observance_shift.

        The holiday may also specify an 'observe_after' such that it is always
        shifted after a preceding holiday. For example, Boxing day is always
        observed after Christmas Day is observed.
        """
        # observance_shift may be overridden in the holiday itself
        shift = getattr(self, 'observance_shift', calendar.observance_shift)
        if callable(shift):
            shifted = shift(self, calendar)
        else:
            shift = shift or {}
            delta = rd.relativedelta(**shift)
            try:
                weekend_days = calendar.get_weekend_days()
            except NotImplementedError:
                weekend_days = ()
            should_shift = self.weekday() in weekend_days
            shifted = self + delta if should_shift else self
        precedent = getattr(self, 'observe_after', None)
        while precedent and shifted <= precedent.get_observed_date(calendar):
            shifted += timedelta(days=1)
        return shifted


class SeriesShiftMixin:
    """
    "Series" holidays like the two Islamic Eid's or Chinese Spring Festival span
    multiple days. If one of these days encounters a non-zero observance_shift,
    apply that shift to all subsequent members of the series.
    """

    def get_calendar_holidays(self, year):
        """
        Ensure that all events are observed in the order indicated.
        """
        days = super().get_calendar_holidays(year)
        holidays = sorted(map(Holiday._from_resolved_definition, days))
        from more_itertools import pairwise
        for a, b in pairwise(holidays):
            b.observe_after = a
        return holidays



from ..registry_tools import iso_register
from .core import UnitedStates


@iso_register('US-OK')
class Oklahoma(UnitedStates):
    """Oklahoma"""
    include_thanksgiving_friday = True
    include_boxing_day = True
    include_columbus_day = False



from datetime import timedelta
from ..core import OrthodoxCalendar
from ..registry_tools import iso_register


@iso_register('BY')
class Belarus(OrthodoxCalendar):
    'Belarus'
    "as of http://president.gov.by/en/holidays_en/"
    # Civil holidays
    include_labour_day = True

    # Christian holidays
    include_christmas = True
    christmas_day_label = "Christmas Day (Catholic)"
    orthodox_christmas_day_label = "Christmas (Orthodox)"

    FIXED_HOLIDAYS = OrthodoxCalendar.FIXED_HOLIDAYS + (
        (1, 2, "Day After New Year"),
        (3, 8, "International Women's Day"),
        (5, 9, "Victory Day"),
        (7, 3, "Republic Day"),
        (11, 7, "October Revolution Day"),
    )
    # Radonitsa
    # https://en.wikipedia.org/wiki/Radonitsa

    def get_radonitsa(self, year):
        # 9 Days after the Orthodox easter sunday
        return self.get_easter_sunday(year) + timedelta(days=9)

    def get_variable_days(self, year):
        days = super().get_variable_days(year)
        days.append((self.get_radonitsa(year), "Radonitsa"))
        return days



from ..core import WesternCalendar, FRI
from ..registry_tools import iso_register


@iso_register('MH')
class MarshallIslands(WesternCalendar):
    "Marshall Islands"
    FIXED_HOLIDAYS = WesternCalendar.FIXED_HOLIDAYS + (
        (3, 3, "Remembrance Day"),
        (5, 1, "Constitution Day"),
        (11, 17, "Presidents' Day"),
        (12, 31, "New Year's Eve"),
    )
    include_good_friday = True

    def get_variable_days(self, year):
        days = super().get_variable_days(year)
        days.append((
            MarshallIslands.get_nth_weekday_in_month(year, 7, FRI),
            "Fishermen's Holiday"
        ))
        days.append((
            MarshallIslands.get_nth_weekday_in_month(year, 9, FRI),
            "Labour Day"
        ))
        days.append((
            MarshallIslands.get_last_weekday_in_month(year, 9, FRI),
            "Manit Day"
        ))
        days.append((
            MarshallIslands.get_nth_weekday_in_month(year, 12, FRI),
            "Gospel Day"
        ))
        return days



from .core import UnitedStates
from ..registry_tools import iso_register


@iso_register('US-KY')
class Kentucky(UnitedStates):
    """Kentucky"""
    include_good_friday = True
    include_thanksgiving_friday = True
    include_christmas_eve = True
    include_columbus_day = False
    include_federal_presidents_day = False
    FIXED_HOLIDAYS = UnitedStates.FIXED_HOLIDAYS + (
        (12, 31, "New Year's Eve"),
    )



from ..core import IslamoWesternCalendar, SAT, SUN
from ..registry_tools import iso_register


@iso_register('CI')
class IvoryCoast(IslamoWesternCalendar):
    "Ivory Coast"
    # Civil holidays
    include_labour_day = True
    # Christian holidays
    include_easter_monday = True
    include_ascension = True
    include_whit_monday = True
    include_assumption = True
    include_all_saints = True
    # Islamic holidays
    include_day_after_prophet_birthday = True
    include_eid_al_fitr = True
    include_day_of_sacrifice = True
    include_day_of_sacrifice_label = "Feast of the Sacrifice"

    FIXED_HOLIDAYS = IslamoWesternCalendar.FIXED_HOLIDAYS + (
        (8, 7, "Independence Day"),
        (11, 15, "National Peace Day"),
    )

    # Ivory Coast has adopted the "western" workweek.
    WEEKEND_DAYS = (SAT, SUN)



from ..registry_tools import iso_register
from .core import UnitedStates


@iso_register('US-TN')
class Tennessee(UnitedStates):
    """Tennessee"""
    include_columbus_day = False

    include_good_friday = True
    include_christmas_eve = True



from ..core import MON
from ..registry_tools import iso_register
from .core import UnitedStates


@iso_register('US-RI')
class RhodeIsland(UnitedStates):
    """Rhode Island"""
    include_federal_presidents_day = False
    include_election_day_even = True

    def get_variable_days(self, year):
        days = super().get_variable_days(year)
        days.append(
            (self.get_nth_weekday_in_month(year, 8, MON, 2), "Victory Day")
        )
        return days
